/*
   Copyright 2013-2015 Skytechnology sp. z o.o.

   This file is part of LizardFS.

   LizardFS is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, version 3.

   LizardFS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with LizardFS. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "common/platform.h"

#include "common/chunk_part_type.h"
#include "unittests/operators.h"

extern const ChunkPartType standard;
extern const ChunkPartType xor_1_of_2;
extern const ChunkPartType xor_2_of_2;
extern const ChunkPartType xor_p_of_2;
extern const ChunkPartType xor_1_of_3;
extern const ChunkPartType xor_2_of_3;
extern const ChunkPartType xor_3_of_3;
extern const ChunkPartType xor_p_of_3;
extern const ChunkPartType xor_1_of_4;
extern const ChunkPartType xor_2_of_4;
extern const ChunkPartType xor_3_of_4;
extern const ChunkPartType xor_4_of_4;
extern const ChunkPartType xor_p_of_4;
extern const ChunkPartType xor_1_of_6;
extern const ChunkPartType xor_2_of_6;
extern const ChunkPartType xor_3_of_6;
extern const ChunkPartType xor_4_of_6;
extern const ChunkPartType xor_5_of_6;
extern const ChunkPartType xor_6_of_6;
extern const ChunkPartType xor_p_of_6;
extern const ChunkPartType xor_1_of_7;
extern const ChunkPartType xor_2_of_7;
extern const ChunkPartType xor_3_of_7;
extern const ChunkPartType xor_4_of_7;
extern const ChunkPartType xor_5_of_7;
extern const ChunkPartType xor_6_of_7;
extern const ChunkPartType xor_7_of_7;
extern const ChunkPartType xor_p_of_7;
extern const ChunkPartType xor_1_of_9;
extern const ChunkPartType xor_2_of_9;
extern const ChunkPartType xor_3_of_9;
extern const ChunkPartType xor_4_of_9;
extern const ChunkPartType xor_5_of_9;
extern const ChunkPartType xor_6_of_9;
extern const ChunkPartType xor_7_of_9;
extern const ChunkPartType xor_8_of_9;
extern const ChunkPartType xor_9_of_9;
extern const ChunkPartType xor_p_of_9;
